/*
 * w32init.c
 *
 * Copyright (C) 2006  Insigme Co., Ltd
 *
 * This software has been developed while working on the Linux Unified Kernel
 * project (http://www.longene.org) in the Insigma Research Institute,  
 * which is a subdivision of Insigma Co., Ltd (http://www.insigma.com.cn).
 * 
 * The project is sponsored by Insigma Co., Ltd.
 *
 * The authors can be reached at linux@insigma.com.cn.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of  the GNU General  Public License as published by the
 * Free Software Foundation; either version 2 of the  License, or (at your
 * option) any later version.
 *
 * Revision History:
 *   Jan 2006 - Created.
 */
 
/*
 * w32init.c: initiliase w32systemcall and some object classes
 */
#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>

#include "mutex.h"
#include "event.h"
#include "semaphore.h"
#include "section.h"
#include "file.h"
#include "thread.h"
#include "process.h"
#include "apc.h"
#include "handle.h"
#include "io.h"

#ifdef CONFIG_UNIFIED_KERNEL

char *rootdir = "/root/.wine/";

extern SSDT_ENTRY KeServiceDescriptorTable[];
asmlinkage int w32system_call(void);
int set_w32system_gate(unsigned int, void *);
int backup_idt_entry(unsigned int n, unsigned long *a, unsigned long *b);
int restore_idt_entry(unsigned int n, unsigned long a, unsigned long b);
unsigned long orig_idt_2e_a, orig_idt_2e_b;
int init_pe_binfmt(void);
void exit_pe_binfmt(void);
#ifdef EXE_SO
int init_exeso_binfmt(void);
void exit_exeso_binfmt(void);
#endif

struct proc_dir_entry *unifiedkernel_proc;
struct list_head object_class_list;

extern int proc_uk_init(void);
extern void proc_uk_exit(void);

extern void kernel_init_registry(void);
extern void init_process_manager(void);
extern void init_section_implement(void);
extern void display_object_dir(POBJECT_DIRECTORY DirectoryObject, LONG Depth);
extern void display_name_info(void);
extern void exit_object(void);
extern void init_named_pipe(void);
extern void init_directories(void);
extern struct task_struct* save_kernel_task;
extern int kthread_stop(struct task_struct*);

/* w32_init */
static int w32_init(void)
{
	ktrace("Unifiedkernel loading...\n");
	/* store the original address that the 0x2E points */ 
	if (backup_idt_entry(0x2E, &orig_idt_2e_a, &orig_idt_2e_b) == -1) {
		kdebug("Module not loaded. backup_idt_entry error: bad idt entry\n");
		return -1;
	}
	
	/* initialize 0x2E */
	if (set_w32system_gate(0x2E, &w32system_call) == -1) {
		kdebug("Module not loaded. set_w32system_gate error: bad idt entry\n");
		return -1;
	}
	
	proc_uk_init();

	/* initialise the internal bits */
	INIT_LIST_HEAD(&object_class_list);
	init_pe_binfmt();
#ifdef EXE_SO
	init_exeso_binfmt();
#endif
	init_handle_tables();
	init_object();
	init_symbol_link();
	init_io();
	init_directories();
	init_named_pipe();
	init_cid_table();
	init_process_manager();
	init_section_implement();
	init_semaphore_implement();
	init_event_implement();
	init_mutant_implement();
	kernel_init_registry();
	display_object_dir(name_space_root, 1);

	register_binfmt(NULL);
	return 0;
} /* end w32_exit */ 
/* w32_exit */
static void w32_exit(void)
{
	int ret;

	destroy_cid_table();
	exit_object();
#ifdef EXE_SO
	exit_exeso_binfmt();
#endif
	exit_pe_binfmt();
	proc_uk_exit();
	ret=wake_up_process(save_kernel_task);
	kthread_stop(save_kernel_task);


	/* restore 0x2E */
	restore_idt_entry(0x2E, orig_idt_2e_a, orig_idt_2e_b);

	ktrace("Module w32 Off!\n");
} /*end w32_exit */ 

module_init(w32_init);
module_exit(w32_exit);
module_param(rootdir, charp, S_IRUGO);
MODULE_LICENSE("GPL");
#endif

