/*
 * thread.c
 *
 * Copyright (C) 2006  Insigme Co., Ltd
 *
 * This software has been developed while working on the Linux Unified Kernel
 * project (http://www.longene.org) in the Insigma Research Institute,
 * which is a subdivision of Insigma Co., Ltd (http://www.insigma.com.cn).
 *
 * The project is sponsored by Insigma Co., Ltd.
 *
 * The authors can be reached at linux@insigma.com.cn.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of  the GNU General  Public License as published by the
 * Free Software Foundation; either version 2 of the  License, or (at your
 * option) any later version.
 *
 * Revision History:
 *   Jan 2006 - Created.
 */

/*
 * thread.c: thread implementation
 * Refered to ReactOS code
 */
#include "win32_process.h"
#include <linux/win32_thread.h>
#include <linux/syscalls.h>
#include <linux/mm.h>
#include <asm/uaccess.h>
#include <asm/ldt.h>
#include <linux/file.h>
#include <linux/nls.h>
#include "apc.h"
#include "file.h"
#include "io.h"
#include "mutex.h"
#include "object.h"
#include "pefile.h"
#include "unistr.h"
#include "attach.h"
#include "event.h"
#include "handle.h"
#include "ke.h"
#include "ntstatus.h"
#include "objwait.h"
#include "process.h"
#include "section.h"
#include "semaphore.h"
#include "virtual.h"
#include "win32.h"
#include "thread.h"
#include "w32syscall.h"

#include "wineserver/uk_lib.h"
#include "wineserver/list.h"
#include "wineserver/user.h"

#ifdef CONFIG_UNIFIED_KERNEL

#ifdef SERVER_PM
static struct list_head w32thread_list = LIST_INIT(w32thread_list);
#endif
POBJECT_TYPE thread_object_type = NULL;
EXPORT_SYMBOL(thread_object_type);

static void thread_close(struct ethread *);
static int  thread_signal(struct ethread *, int);
static void thread_exit(struct ethread *, int);
static void thread_execve(struct ethread *);
static void thread_fork(struct ethread *,
		struct task_struct *,
		struct task_struct *,
		unsigned long);

extern void do_exit_task(struct task_struct *tsk, long code);
extern long do_fork_from_task(struct task_struct *ptsk,
		unsigned long process_flags,
		unsigned long clone_flags,
		unsigned long stack_start,
		struct pt_regs *regs,
		unsigned long stack_size,
		int __user *parent_tidptr,
		int __user *child_tidptr);

extern asmlinkage void thread_startup(void);

extern unsigned long get_ntdll_entry(void);
extern unsigned long get_interp_entry(void);
extern unsigned long get_thread_entry(void);

#define arch_init_thread        i386_init_thread
#define i386_init_thread(th, ctx)       do { } while (0)

static const struct ethread_operations ethread_ops = {
		name:           "ethread",
		owner:          THIS_MODULE,
		close:          thread_close,
		exit:           thread_exit,
		signal:         thread_signal,
		execve:         thread_execve,
		fork:           thread_fork
};

void __attribute__((regparm(3)))
set_trap_frame(unsigned long esp, struct task_struct *tsk)
{
	if (tsk->ethread)
		tsk->ethread->tcb.trap_frame = (struct ktrap_frame *)esp;
}

int set_tls_array(struct thread_struct *t, int idx, unsigned long addr, unsigned int limit)
{
	struct user_desc info;
	struct desc_struct *desc;

	if (idx < GDT_ENTRY_TLS_MIN || idx > GDT_ENTRY_TLS_MAX)
		return -EINVAL;

	desc = t->tls_array + idx - GDT_ENTRY_TLS_MIN;

	info.entry_number = idx;
	info.base_addr = addr;
	info.limit = limit;
	info.contents = 0;
	info.read_exec_only = 0;
	info.seg_not_present = 0;
	info.seg_32bit = 1;
	info.limit_in_pages = 1;
	info.useable = 0;

/* FIXME */
	fill_ldt(desc, &info);
/*
	desc->a = LDT_entry_a(&info);
	desc->b = LDT_entry_b(&info);
*/

	return 0;
}

#ifdef SERVER_PM
/* copy a context structure according to the flags */
void copy_context( CONTEXT *to, const CONTEXT *from, unsigned int flags )
{
    flags &= ~CONTEXT_i386;  /* get rid of CPU id */
    if (flags & CONTEXT_CONTROL) {
        to->Ebp    = from->Ebp;
        to->Eip    = from->Eip;
        to->Esp    = from->Esp;
        to->SegCs  = from->SegCs;
        to->SegSs  = from->SegSs;
        to->EFlags = from->EFlags;
    }
    if (flags & CONTEXT_INTEGER) {
        to->Eax = from->Eax;
        to->Ebx = from->Ebx;
        to->Ecx = from->Ecx;
        to->Edx = from->Edx;
        to->Esi = from->Esi;
        to->Edi = from->Edi;
    }
    if (flags & CONTEXT_SEGMENTS) {
        to->SegDs = from->SegDs;
        to->SegEs = from->SegEs;
        to->SegFs = from->SegFs;
        to->SegGs = from->SegGs;
    }
    if (flags & CONTEXT_FLOATING_POINT) {
        to->FloatSave = from->FloatSave;
    }
    if (flags & CONTEXT_EXTENDED_REGISTERS) {
        memcpy( to->ExtendedRegisters, from->ExtendedRegisters, sizeof(to->ExtendedRegisters) );
    }
    if (flags & CONTEXT_DEBUG_REGISTERS) {
        to->Dr0 = from->Dr0;
        to->Dr1 = from->Dr1;
        to->Dr2 = from->Dr2;
        to->Dr3 = from->Dr3;
        to->Dr6 = from->Dr6;
        to->Dr7 = from->Dr7;
    }
    to->ContextFlags |= flags;
}

/* retrieve the current instruction pointer of a context */
static inline void *get_context_ip( const CONTEXT *context )
{
    return (void *)context->Eip;
}

/* return the context flag that contains the CPU id */
static inline unsigned int get_context_cpu_flag(void)
{
    return CONTEXT_i386;
}

/* return only the context flags that correspond to system regs */
/* (system regs are the ones we can't access on the client side) */
static inline unsigned int get_context_system_regs( unsigned int flags )
{
    return flags & (CONTEXT_DEBUG_REGISTERS & ~CONTEXT_i386);
}

/* Temporarily add just for compiling 
 * These functions shouldn't be here */
void set_thread_context( struct w32thread *thread, const CONTEXT *context, unsigned int flags )
{
}
void get_thread_context( struct w32thread *thread, CONTEXT *context, unsigned int flags )
{
}
void security_set_thread_token(struct w32thread *thread, obj_handle_t handle)
{
}
void remove_process_thread( struct w32process *process, struct w32thread *thread )
{
}
/* Temporarily add just for compiling */

/* initialize the structure for a newly allocated w32thread */
static inline void init_w32thread_structure( struct w32thread *thread )
{
    thread->debug_ctx       = NULL;
    thread->debug_event     = NULL;
    thread->debug_break     = 0;
    thread->queue           = NULL;
    thread->error           = 0;
    thread->req_data        = NULL;
    thread->req_toread      = 0;
    thread->reply_data      = NULL;
    thread->reply_towrite   = 0;
    thread->state           = RUNNING;
    thread->context         = NULL;
    thread->suspend_context = NULL;
    thread->desktop_users   = 0;

    thread->creation_time = current_time;
    thread->exit_time     = 0;
}

static void cleanup_thread( struct w32thread *thread )
{
    kfree( thread->req_data );
    kfree( thread->reply_data );
    kfree( thread->suspend_context );
    free_msg_queue( thread );
    cleanup_clipboard_thread(thread);
    destroy_thread_windows( thread );
    close_thread_desktop( thread );
    thread->req_data = NULL;
    thread->reply_data = NULL;
    thread->context = NULL;
    thread->suspend_context = NULL;
    thread->desktop = 0;
}

static inline void free_w32thread(struct w32thread *thread)
{
	ktrace("%s\n", __FUNCTION__);
	cleanup_thread(thread);
	kfree(thread);
}

/* create a new w32thread */
struct w32thread *create_w32thread( struct ethread *ethread )
{
	struct w32thread *w32thread;
	struct w32process *w32process = ethread->threads_process->win32process;

	ktrace("%s\n", __FUNCTION__);
	w32thread = mem_alloc(sizeof(struct w32thread));
	if (!w32thread)
		return NULL;

	init_w32thread_structure( w32thread );

	w32thread->process = w32process;
	w32thread->ethread = ethread;
	w32thread->desktop = w32process->desktop;
	if (!current_thread) 
		current_thread = w32thread;

	list_add_head( &w32thread_list, &w32thread->entry );

	return w32thread;
}

void kill_w32thread( struct w32thread *thread )
{
    if (thread->state == TERMINATED)
        return;  /* already killed */
    thread->state = TERMINATED;
    thread->exit_time = current_time;
    if (current_thread == thread)
        current_thread = NULL;
    
    kill_console_processes( thread, 0 );
    cleanup_thread( thread );
    remove_process_thread( thread->process, thread );
}

/* terminate a thread */
NTSTATUS server_terminate_thread(struct ethread *ethread)
{
    struct w32thread *thread;

    if (!ethread)
	    return STATUS_INVALID_PARAMETER;

    thread = ethread->tcb.win32thread;
    if (thread != current_thread) 
	    kill_w32thread( thread );
    return STATUS_SUCCESS;
}

/* retrieve the current context of a thread */
int server_get_thread_context(struct ethread *ethread, int suspend, unsigned int context_flags)
{
    struct w32thread *thread = ethread->tcb.win32thread;
    CONTEXT *context;

    if (suspend) {
        if (thread != current_thread || !thread->suspend_context)
            set_error( STATUS_INVALID_PARAMETER ); /* not suspended, shouldn't happen */
        else {
            if (thread->context == thread->suspend_context)
                thread->context = NULL;
            set_reply_data_ptr( thread->suspend_context, sizeof(CONTEXT) );
            thread->suspend_context = NULL;
        }
    } else if (thread != current_thread && !thread->context) {
        /* thread is not suspended, retry (if it's still running) */
        if (thread->state != RUNNING)
            set_error( STATUS_ACCESS_DENIED );
        else
            set_error( STATUS_PENDING );
    }
    else if ((context = set_reply_data_size( sizeof(CONTEXT) ))) {
        unsigned int flags = get_context_system_regs( context_flags );

        memset( context, 0, sizeof(CONTEXT) );
        context->ContextFlags = get_context_cpu_flag();
        if (thread->context)
            copy_context( context, thread->context, context_flags & ~flags );
        if (flags)
            get_thread_context( thread, context, flags );
    }
    return (thread == current_thread);
}

/* set the current context of a thread */
int server_set_thread_context(struct ethread *ethread, int suspend, unsigned int context_flags)
{
    struct w32thread *thread = ethread->tcb.win32thread;

    if (suspend) {
        if (thread != current_thread || thread->context)
            set_error( STATUS_INVALID_PARAMETER ); /* nested suspend or exception, shouldn't happen */
        else if ((thread->suspend_context = mem_alloc( sizeof(CONTEXT) ))) {
            memcpy( thread->suspend_context, get_req_data(), sizeof(CONTEXT) );
            thread->context = thread->suspend_context;
        }
    }
    else if (thread != current_thread && !thread->context) {
        /* thread is not suspended, retry (if it's still running) */
        if (thread->state != RUNNING)
            set_error( STATUS_ACCESS_DENIED );
        else
            set_error( STATUS_PENDING );
    }
    else {
        const CONTEXT *context = get_req_data();
        unsigned int flags = get_context_system_regs( context_flags );

        if (flags)
            set_thread_context( thread, context, flags );
        if (thread->context && !get_error())
            copy_context( thread->context, context, context_flags & ~flags );
    }
    return (thread == current_thread);
}
#endif

/*
 * thread object destructor
 */     
static void thread_close(struct ethread *thread)
{       
	struct eprocess *process;

	process = thread->threads_process;
	thread->threads_process = NULL;

	/* detach the thread record from the Linux task */
	ktrace("thread_close, obj %p\n", thread);
	thread->et_task = NULL;
	thread->tcb.state = Terminated;

	deref_object((PVOID)thread);

	/* detach from the containing process */
	deref_object(process);
} /* end thread_close() */

void rundown_thread(void)
{
	struct kthread * thread = (struct kthread *) get_current_ethread();
	struct kmutant * mutant;
	struct list_head * cur_entry;

	ktrace("rundown thread\n");

	local_irq_disable();
	while (!list_empty(&thread->mutant_list_head)) {
		/* Get the Mutant */
		cur_entry = thread->mutant_list_head.next;
		mutant = list_entry(cur_entry, struct kmutant, mutant_list_entry);

		/* check apc disable */

		mutant->header.signal_state = 1;
		mutant->abandoned = 1;
		mutant->owner_thread = NULL;
		list_del(&mutant->mutant_list_entry);

		if(!list_empty(&mutant->header.wait_list_head)) {
			wait_test(&mutant->header, MUTANT_INCREMENT);
		}
	}
	local_irq_enable();
}

/*
 * notification of exit
 * - the exit_status is as sys_wait() would return
 * - notification includes fatal signals
 */
static void thread_exit(struct ethread *thread, int exit_status)
{
	struct eprocess	*process = thread->threads_process;
	BOOLEAN last;

	/* if Terminated, do nothing */
	ktrace("thread %p, exit_status %d\n", thread, thread->exit_status);

	thread->terminated = 1;

	/* Can't terminate a thread if it attached another process */
	if (thread->tcb.apc_state_index) {
		return;
	}

	/* TODO: Lower to Passive Level */
	
	/* Lock the Process before we modify its thread entries */
	lock_process(process);

	list_del(&thread->thread_list_entry);

	/* TODO: close port */
	
	/* TODO: Rundown Win32 Structures */

    	/* Set the last Thread Exit Status */
	process->last_thread_exit_status = thread->exit_status;
	
	/* The last Thread shuts down the Process */
	if ((last = list_empty(&process->thread_list_head))) {
		/* Save the Exit Time if not already done by NtTerminateProcess. This 
		   happens when the last thread just terminates without explicitly 
		   terminating the process. TODO */
#if 0
		process->exit_time = thread->exit_time;
#endif
		__exit_process(process);
	}

	if (thread->tcb.win32thread) {
#ifndef SERVER_PM
		kfree(thread->tcb.win32thread);
#else
		free_w32thread(thread->tcb.win32thread);
#endif
		thread->tcb.win32thread = NULL;
	}

	/* Free the TEB, if last thread, teb is freed by exit() */
	if (thread->tcb.teb && !last) {
		delete_teb(thread->tcb.teb);
		thread->tcb.teb = NULL;
	}

	list_del(&thread->tcb.thread_list_entry);

	/* Unlock the Process */		
	unlock_process(process);

	/* Rundown Mutexes */
	rundown_thread();

	/* Satisfy waits */
	local_irq_disable();
	thread->tcb.header.signal_state = true;
	if (!list_empty(&thread->tcb.header.wait_list_head))
		wait_test((struct dispatcher_header *)&thread->tcb, IO_NO_INCREMENT);
	local_irq_enable();

	ktrace("exit thread done\n");
} /* end thread_exit() */

/*
 * notification of signal
 */
static int thread_signal(struct ethread *thread, int signal)
{
	return WIN32_THREAD_SIGNAL_OKAY;
} /* end thread_signal() */

/*
 * notification of execve
 * - if this is NULL, a thread object will be destroyed on execve
 */
static void thread_execve(struct ethread *thread)
{       
	/* TODO */
} /* end thread_execve() */

/*
 * notification that fork/clone has set up the new process and
 * is just about to dispatch it
 * - no threads will have been copied by default
 */
static void thread_fork(struct ethread *thread,
		struct task_struct *parent,
		struct task_struct *child,
		unsigned long clone_flags)
{       
	/* TODO */
} /* end thread_fork() */

VOID delete_thread(PVOID Object)
{
	struct ethread * thread = Object;

	ktrace("delete_thread, obj %p\n", Object);

	if (thread->cid.unique_thread)
		delete_cid_handle(thread->cid.unique_thread, thread_object_type);

	/* TODO: release stack mem */
}

/* FIXME */
int poll_thread(struct wait_table_entry *wte)
{
	struct ethread * thread = (struct ethread * )wte->wte_obj;
	int ret;

	ret = (thread->tcb.state == Running) ? POLL_NOTSIG : POLL_SIG;

	return ret;
}

/*      
 * set teb on fs
 */     
int set_teb_selector(struct task_struct *tsk, long teb)
{
	int	cpu;

	set_tls_array(&tsk->thread, TEB_SELECTOR >> 3, teb, 1);
	cpu = get_cpu();
	load_TLS(&tsk->thread, cpu);
	put_cpu();

	return 0;
} /* end set_teb_selector */

/* initialize kthread */
void kthread_init(struct kthread *thread, struct eprocess *process)
{
	ktrace("kthread_init\n");

	INIT_DISP_HEADER(&thread->header, thread_object, sizeof(struct ethread), false);

	/* initialize the mutant list */
	INIT_LIST_HEAD(&thread->mutant_list_head);

	/* setup apc fields */
	INIT_LIST_HEAD(&thread->apc_state.apc_list_head[0]);
	INIT_LIST_HEAD(&thread->apc_state.apc_list_head[1]);
	INIT_LIST_HEAD(&thread->saved_apc_state.apc_list_head[0]);
	INIT_LIST_HEAD(&thread->saved_apc_state.apc_list_head[1]);
	thread->apc_state.process = (struct kprocess *)process;
	thread->apc_state_pointer[OriginalApcEnvironment] = &thread->apc_state;
	thread->apc_state_pointer[AttachedApcEnvironment] = &thread->saved_apc_state;
	thread->apc_state_index = OriginalApcEnvironment;
	thread->apc_queue_lock = SPIN_LOCK_UNLOCKED;
	thread->apc_queueable = true;

	/*NOW FIXME Initialize the Suspend APC */
	apc_init(&thread->suspend_apc, 
			thread, 
			OriginalApcEnvironment, 
			suspend_thread_kernel_routine, 
			NULL, 
			suspend_thread_normal_routine, 
			KernelMode, 
			NULL);


	/* Initialize the Suspend Semaphore */
	semaphore_init(&thread->suspend_semaphore, 0, 128);

	/* initialize the suspend semaphore */
	/* FIXME: sema_init(&thread->suspend_semaphore, 0); */
	/* FIXME: keinitializetimer(&thread->timer); */

	arch_init_thread(thread, context);

	thread->base_priority = process->pcb.base_priority;
	thread->quantum = process->pcb.quantum_reset;
	thread->quantum_reset = process->pcb.quantum_reset;
	thread->affinity = process->pcb.affinity;
	thread->priority = process->pcb.base_priority;
	thread->user_affinity = process->pcb.affinity;
	thread->disable_boost = process->pcb.disable_boost;
	thread->auto_alignment = process->pcb.auto_alignment;

	/* set the thread to initalized */
	thread->state = Initialized;

	lock_process(process);
	list_add_tail(&thread->thread_list_entry, &process->pcb.thread_list_head);
	unlock_process(process);
#ifdef SERVER_PM
	thread->win32thread = create_w32thread((struct ethread *)thread);
#endif
} /* end kthread_init */

/* initialize ethread */
void ethread_init(struct ethread *thread, struct eprocess *process, struct task_struct *tsk)
{
	ktrace("ethread_init\n");

	/* attach to the containing process */
	ref_object((PVOID)process);
	write_lock(&process->ep_lock);
	thread->threads_process = process;
	write_unlock(&process->ep_lock);

	/* FIXME create a thread object and hook in to the Linux task */
	thread->et_task = tsk;

	atomic_set(&thread->et_count, 0);

	/* FIXME */
	thread->et_ops = (struct ethread_operations *)&ethread_ops;

	INIT_LIST_HEAD(&thread->lpc_reply_chain);
	INIT_LIST_HEAD(&thread->irp_list);
	INIT_LIST_HEAD(&thread->active_timer_list_head);
	thread->active_timer_list_lock = SPIN_LOCK_UNLOCKED;
	thread->thread_lock = SPIN_LOCK_UNLOCKED;

	/* FIXME: semaphore_init */

	thread->cid.unique_process = process->unique_processid;

	thread->win32_start_address = 0;        /* context->Eax, default is 0 */
	lock_process(process);
	list_add_tail(&thread->thread_list_entry, &process->thread_list_head);
	unlock_process(process);

	add_ethread(thread->et_task, thread);
	if (atomic_read(&thread->et_count) == 1)	/* FIXME: add this to win32_thread.c */
		ref_object(thread);

#ifndef SERVER_PM
	if ((thread->tcb.win32thread = kmalloc(sizeof(struct w32thread), GFP_KERNEL)))
		memset(thread->tcb.win32thread, 0, sizeof(struct w32thread));
#endif

	kthread_init(&thread->tcb, process);
} /* end ethread_init */

/*
 * user_thread_startup
 * prepare for jumping to userspace to execute after new thread waken
 */
VOID
STDCALL
user_thread_startup(PKSTART_ROUTINE StartRoutine,
		PVOID StartContext)
{
	struct ethread  *thread;
	PKAPC   thread_apc;
	void * start_stack;

	ktrace("user_thread_startup, pid %x, tgid %x\n",current->pid, current->tgid);

	if (!(thread = get_current_ethread())) {
		return;
	}

	if (!(thread_apc = kmalloc(sizeof(struct kapc),GFP_KERNEL))) {
		return;
	}

	start_stack = (void *)thread->tcb.stack_base; /* user stack base */

	if (thread->threads_process->fork_in_progress) {
		apc_init(thread_apc,
				&thread->tcb,
				OriginalApcEnvironment,
				thread_special_apc,
				NULL,
				(PKNORMAL_ROUTINE)get_ntdll_entry(),
				UserMode,
				start_stack);
		insert_queue_apc(thread_apc, 
				(void *)get_interp_entry(), 
				thread->threads_process->spare0[0], 
				IO_NO_INCREMENT);

		thread->threads_process->fork_in_progress = NULL;
	} else {
		apc_init(thread_apc,
				&thread->tcb,
				OriginalApcEnvironment,
				thread_special_apc,
				NULL,
				(PKNORMAL_ROUTINE)get_thread_entry(),
				UserMode,
				start_stack);
		insert_queue_apc(thread_apc, NULL,NULL, IO_NO_INCREMENT);
	}
	thread->tcb.apc_state.uapc_pending = 1;
	set_tsk_thread_flag(current, TIF_APC);

	try_module_get(THIS_MODULE);

	return;
} /* end user_thread_startup */

/*
 * init_thread_with_context
 * init context for thread
 */
VOID
STDCALL
init_thread_with_context(struct kthread* Thread,
		PKSYSTEM_ROUTINE SystemRoutine,
		PKSTART_ROUTINE StartRoutine, /* FIXME */
		PVOID StartContext,           /* FIXME */
		PCONTEXT Context)
{
	struct thread_info *info;
	struct task_struct *p;
	unsigned long * trapframe;
	struct pt_regs * regs;
	PCONTEXT context;

	if (!(context = kmalloc(sizeof(*context),GFP_KERNEL))) {
		return;
	}

	if (copy_from_user(context, Context, sizeof(*context))) {
		kfree(context);
		return;
	}

	info = (struct thread_info *)((unsigned long)Thread->kernel_stack- THREAD_SIZE);
	p = info->task;
	set_tls_array(&p->thread, TEB_SELECTOR >> 3, (unsigned long)Thread->teb, 1);

	/* set for switch */
	trapframe = (unsigned long *)((unsigned long)Thread->kernel_stack - 8 - sizeof(struct pt_regs) - 12);
	regs = (struct pt_regs *)(trapframe + 3);

	Thread->trap_frame = (struct ktrap_frame *)regs;
	
	trapframe[0] = (unsigned long)SystemRoutine;
	trapframe[1] = (unsigned long)StartRoutine;
	trapframe[2] = (unsigned long)StartContext;

	p->thread.ip = (unsigned long)thread_startup;
	p->thread.sp = (unsigned long)trapframe;
	p->thread.sp0 = (unsigned long)(regs+1); 
	p->thread.fs = TEB_SELECTOR;

	/* set for userspace */
	regs->sp = context->Esp;
	regs->ip = context->Eip;

	kfree(context);
	return;
} /* end init_thread_with_context */

/*
 * initialize_thread
 * initialize kthread
 */
VOID
STDCALL
initialize_thread(struct kprocess* Process,
		struct kthread* Thread,
		PKSYSTEM_ROUTINE SystemRoutine,
		PKSTART_ROUTINE StartRoutine,  /* FIXME */
		PVOID StartContext,            /* FIXME */
		PCONTEXT Context,
		PVOID Teb,
		PVOID KernelStack)
{
	ktrace("initialize_thread kprocess %p, kthread %p, systemroutine %p, context %p, teb %p, kernelstack %p\n", 
			Process, Thread, SystemRoutine, Context, Teb, KernelStack);

	/* Set the TEB */
	Thread->teb = Teb;

	/* Allocate Stack use linux task stack and init
	 * Set the Thread Stacks
	 */
	Thread->kernel_stack = (PCHAR)KernelStack + THREAD_SIZE;

	/* FIXME
	 * Establish the pde's for the new stack and the thread structure within the
	 * address space of the new process. They are accessed while taskswitching or
	 * while handling page faults. At this point it isn't possible to call the
	 * page fault handler for the missing pde's.
	 */
	update_page_dir(((struct ethread *)Thread)->et_task->mm, (void *)Thread->stack_limit, THREAD_SIZE);
	update_page_dir(((struct ethread *)Thread)->et_task->mm, Thread, sizeof(struct ethread));

	init_thread_with_context(Thread,
			SystemRoutine,
			StartRoutine,
			StartContext,
			Context);
} /* end initialize_thread */

/*
 * create_thread
 */
NTSTATUS
STDCALL
create_thread(OUT PHANDLE ThreadHandle,
		IN ACCESS_MASK DesiredAccess,
		IN POBJECT_ATTRIBUTES ObjectAttributes  OPTIONAL,
		IN HANDLE ProcessHandle,
		IN struct eprocess* TargetProcess,   			/* FIXME */
		OUT PCLIENT_ID ClientId,
		IN PCONTEXT ThreadContext,
		IN PINITIAL_TEB InitialTeb,
		IN BOOLEAN CreateSuspended,
		IN PKSTART_ROUTINE StartRoutine OPTIONAL,   	/* FIXME */
		IN PVOID StartContext OPTIONAL)            	/* FIXME */
{
	struct eprocess * process;
	struct ethread * thread, *first_thread, *cur_thread;
	struct task_struct *new_tsk = NULL;
	unsigned clone_flags = 0;
	PTEB teb_base;
	long   cpid;
	HANDLE hthread;
	NTSTATUS status = STATUS_SUCCESS;

	ktrace("create_thread\n");

	if (!(cur_thread = get_current_ethread())) {
		return STATUS_INVALID_PARAMETER;
	}

	/* current still be regarded */
	if (ProcessHandle && ProcessHandle != NtCurrentProcess()) {
		status = ref_object_by_handle(ProcessHandle,
				PROCESS_ALL_ACCESS,
				process_object_type,
				KernelMode,
				(PVOID *)&process,
				NULL);
		if (!NT_SUCCESS(status))
			return status;
	} else {
		if (TargetProcess)
			process = (struct eprocess * )TargetProcess;
		else
			process = cur_thread->threads_process;
		ref_object(process);
	}

	if (!process->fork_in_progress) {
		/* second and after */
		if (!ProcessHandle || ProcessHandle == NtCurrentProcess())
			first_thread = cur_thread;
		else
			first_thread = get_first_thread(process);
		if (!first_thread) {
			status = STATUS_INVALID_PARAMETER;
			goto cleanup_process;
		}
		
		clone_flags = SIGCHLD | CLONE_VM | CLONE_FS | CLONE_FILES | CLONE_SIGHAND | CLONE_PARENT | CLONE_THREAD;
		cpid = do_fork_from_task(first_thread->et_task, CREATE_THREAD, clone_flags,
				first_thread->tcb.trap_frame->esp, (struct pt_regs *)first_thread->tcb.trap_frame,
			       	0, NULL, NULL);
		if (cpid < 0) {
			status = STATUS_INVALID_PARAMETER;
			goto cleanup_process;
		}

		new_tsk = find_task_by_vpid(cpid);
		
		memset(&new_tsk->thread.tls_array, 0, sizeof(new_tsk->thread.tls_array));
		set_tls_array(&new_tsk->thread, first_thread->et_task->thread.gs >> 3, 
				(unsigned long)InitialTeb->StackBase + 0x800, 0xfffff);

		/* allocate a Win32 thread object */
		status = create_object(KernelMode,
				thread_object_type,
				ObjectAttributes,
				KernelMode,
				NULL,
				sizeof(struct ethread),
				0,
				0,
				(PVOID *)&thread);
		if (!NT_SUCCESS(status))
			goto cleanup_tsk;

		ethread_init(thread, process, new_tsk);
		deref_object(thread);

	} else {
		/* for first thread */
		thread = process->fork_in_progress;
		new_tsk = thread->et_task;
	}

	thread->cid.unique_thread = create_cid_handle(thread, thread_object_type);
	thread->cid.unique_process = process->unique_processid;
	if (!thread->cid.unique_thread) {
		goto cleanup_tsk;
	}

	if (ClientId) {
		if (copy_to_user(&ClientId->UniqueThread, &thread->cid.unique_thread, sizeof(HANDLE)))
			goto cleanup_tsk;
		if (copy_to_user(&ClientId->UniqueProcess, &thread->cid.unique_process, sizeof(HANDLE)))
			goto cleanup_tsk;
	}

	/* set user stack base */
	thread->tcb.stack_base = InitialTeb->StackBase;

	if (ThreadContext) {
		/* Create Teb */
		teb_base = create_teb(process, (PCLIENT_ID)&thread->cid, InitialTeb);

		/* Set the Start Addresses */
		thread->start_address = (PVOID)ThreadContext->Eip;
		thread->win32_start_address = (PVOID)ThreadContext->Eax; /*FIXME */

		/* intialize kthread */
		initialize_thread(&process->pcb,
				&thread->tcb,
				user_thread_startup,
				NULL,
				NULL,
				ThreadContext,
				teb_base,
				task_thread_info(new_tsk));
	} else {
		/* FIXME PsCreateSystemThread */
	}

	/* Insert Thread into Handle Table */
	status = insert_object(thread,
			NULL,
			DesiredAccess,
			0,
			NULL,
			&hthread);
	if (!NT_SUCCESS(status))
		goto cleanup_tsk;

	if (copy_to_user(ThreadHandle, &hthread, sizeof(hthread))) {
		status = STATUS_UNSUCCESSFUL;
		goto cleanup_thread;
	}

	clear_tsk_need_resched(new_tsk);
	sched_fork(new_tsk, clone_flags);
	wake_up_new_task(new_tsk, CLONE_VM | CLONE_FS | CLONE_FILES| CLONE_SIGHAND);

	/* FIXME Notify Thread Creation */

	/* NOW FIXME Suspend the Thread if we have to */
	if (CreateSuspended)
		suspend_thread(&thread->tcb);

	/* FIXME: SECURITY */

	/* FIXME Dispatch thread */
	status = STATUS_SUCCESS;
	goto cleanup_process;

cleanup_thread:
	deref_object(thread);

cleanup_tsk:
	if (new_tsk)
		do_exit_task(new_tsk,0);

cleanup_process:
	deref_object(process);

	return status;
} /* end create_thread */

/*
 * NtCreateThread
 */
NTSTATUS 
SERVICECALL
NtCreateThread(OUT PHANDLE              ThreadHandle,
		IN  ACCESS_MASK          DesiredAccess,
		IN  POBJECT_ATTRIBUTES   ObjectAttributes  OPTIONAL,
		IN  HANDLE               ProcessHandle,
		OUT PCLIENT_ID           ClientId,
		IN  PCONTEXT             ThreadContext,
		IN  PINITIAL_TEB         InitialTeb,
		IN  BOOLEAN              CreateSuspended)
{
	PINITIAL_TEB safe_initial_teb;
	NTSTATUS	status;

	ktrace("NtCreateThread\n");

	if (!(safe_initial_teb = kmalloc(sizeof(INITIAL_TEB), GFP_KERNEL))) {
		return STATUS_NO_MEMORY;
	}

	status = STATUS_UNSUCCESSFUL;
	if (copy_from_user(safe_initial_teb, InitialTeb, sizeof(INITIAL_TEB)))
		goto out;

	/* Call the shared function */
	status = create_thread(ThreadHandle,
			DesiredAccess,
			ObjectAttributes,
			ProcessHandle,
			NULL,
			ClientId,
			ThreadContext,
			safe_initial_teb,
			CreateSuspended,
			NULL,
			NULL);

out:
	kfree(safe_initial_teb);	
	return status;
} /* end NtCreateThread */
EXPORT_SYMBOL(NtCreateThread);

NTSTATUS
SERVICECALL
NtOpenThread(OUT PHANDLE ThreadHandle,
		IN ACCESS_MASK DesiredAccess,
		IN POBJECT_ATTRIBUTES ObjectAttributes  OPTIONAL,
		IN PCLIENT_ID ClientId  OPTIONAL)
{
	KPROCESSOR_MODE pre_mode = (KPROCESSOR_MODE)get_pre_mode();
	struct ethread *thread = NULL;
	NTSTATUS status = STATUS_INVALID_PARAMETER;

	ktrace("NtOpenThread\n");
	if (ObjectAttributes->ObjectName) {
		return open_object_by_name(ObjectAttributes, 
				thread_object_type, 
				NULL, 
				pre_mode, 
				DesiredAccess, 
				NULL, 
				ThreadHandle);
	} else if (ClientId && ClientId->UniqueThread) {
		if ((status = lookup_thread_by_tid(ClientId->UniqueThread, &thread)))
			return status;

		status = open_object_by_pointer(thread, 
				ObjectAttributes->Attributes, 
				NULL, 
				DesiredAccess, 
				thread_object_type, 
				pre_mode,
				ThreadHandle);

		deref_object(thread);
	}

	return status;
}
EXPORT_SYMBOL(NtOpenThread);

void trapframe_to_context(struct ethread * thread, PCONTEXT context)
{
	struct pt_regs * trap_frame =  (struct pt_regs * )thread->tcb.trap_frame; 

	if ((context->ContextFlags & CONTEXT_CONTROL ) == CONTEXT_CONTROL) {
		context->Esp = trap_frame->sp;
		context->SegSs = trap_frame->ss;
		context->SegCs = trap_frame->cs;
		context->Eip = trap_frame->ip;
		context->EFlags = trap_frame->flags;
		context->Ebp = trap_frame->bp;
	} 


	if ((context->ContextFlags & CONTEXT_INTEGER) == CONTEXT_INTEGER) {
		context->Eax = trap_frame->ax;
		context->Ebx = trap_frame->bx;
		context->Ecx = trap_frame->cx;
		context->Edx = trap_frame->dx;
		context->Esi = trap_frame->si;
		context->Edi = trap_frame->di;
	}

	if ((context->ContextFlags & CONTEXT_SEGMENTS ) == CONTEXT_SEGMENTS) {
		context->SegDs = trap_frame->ds;
		context->SegEs = trap_frame->es;
		context->SegFs = thread->et_task->thread.fs;
		context->SegGs = thread->et_task->thread.gs;
	}

	if ((context->ContextFlags & CONTEXT_DEBUG_REGISTERS) == CONTEXT_DEBUG_REGISTERS) {
		context->Dr0 = thread->et_task->thread.debugreg0;
		context->Dr1 = thread->et_task->thread.debugreg1;
		context->Dr2 = thread->et_task->thread.debugreg2;
		context->Dr3 = thread->et_task->thread.debugreg3;
		context->Dr6 = thread->et_task->thread.debugreg6;
		context->Dr7 = thread->et_task->thread.debugreg7;
	}
}

void context_to_trapframe(PCONTEXT context , struct ethread * thread)
{
	struct pt_regs * trap_frame =  (struct pt_regs * )thread->tcb.trap_frame; 

	if ((context->ContextFlags & CONTEXT_CONTROL) == CONTEXT_CONTROL) {
		trap_frame->sp = context->Esp;
		trap_frame->ss = context->SegSs;
		trap_frame->cs = context->SegCs;
		trap_frame->ip = context->Eip;
		trap_frame->flags = context->EFlags;
		trap_frame->bp = context->Ebp;
	} 

	if ((context->ContextFlags & CONTEXT_INTEGER) == CONTEXT_INTEGER) {
		trap_frame->ax = context->Eax;
		trap_frame->bx = context->Ebx;
		trap_frame->cx = context->Ecx;
		trap_frame->dx = context->Edx;
		trap_frame->si = context->Esi;
		trap_frame->di = context->Edi;
	}

	if ((context->ContextFlags & CONTEXT_SEGMENTS) == CONTEXT_SEGMENTS) {
		trap_frame->ds = context->SegDs;
		trap_frame->es = context->SegEs;
		thread->et_task->thread.fs = context->SegFs;
		thread->et_task->thread.gs = context->SegGs;
	}	

	if ((context->ContextFlags & CONTEXT_DEBUG_REGISTERS) == CONTEXT_DEBUG_REGISTERS) {
		thread->et_task->thread.debugreg0 = context->Dr0;
		thread->et_task->thread.debugreg1 = context->Dr1;
		thread->et_task->thread.debugreg2 = context->Dr2;
		thread->et_task->thread.debugreg3 = context->Dr3;
		thread->et_task->thread.debugreg6 = context->Dr6;
		thread->et_task->thread.debugreg7 = context->Dr7;
	}

	/* TODO: float register */
}

VOID
STDCALL
get_set_kernel_context_routine(PKAPC Apc,
                             PKNORMAL_ROUTINE* NormalRoutine,
                             PVOID* NormalContext,
                             PVOID* SystemArgument1,
                             PVOID* SystemArgument2)
{
	PGET_SET_CTX_CONTEXT get_set_context;
	struct kevent * event;
	PCONTEXT context;

	get_set_context = (PGET_SET_CTX_CONTEXT) Apc;
	event = &get_set_context->event;
	context = &get_set_context->context;

	if (*SystemArgument1)
		trapframe_to_context(get_current_ethread(), context);
	else
		context_to_trapframe(context, get_current_ethread());

	set_event(event, EVENT_INCREMENT, FALSE);
	suspend_thread(&current->ethread->tcb);
}

NTSTATUS SERVICECALL
NtGetContextThread(IN  HANDLE   ThreadHandle,
                   OUT PCONTEXT ThreadContext)
{
	struct ethread * thread;
	GET_SET_CTX_CONTEXT	ctx_context;
	NTSTATUS status = STATUS_SUCCESS;

	ktrace("NtGetContextThread %p, %p\n",ThreadHandle, ThreadContext);

	status = ref_object_by_handle(ThreadHandle,
			THREAD_ALL_ACCESS,
			thread_object_type,
			KernelMode,
			(PVOID *)&thread,
			NULL);
	if (!NT_SUCCESS(status)) {
		return status;
	}

	if (thread == get_current_ethread()) {
		trapframe_to_context(thread, &ctx_context.context);
	} else {
		event_init(&ctx_context.event, notification_event, 0);

		apc_init(&ctx_context.apc, 
				&thread->tcb, 
				OriginalApcEnvironment, 
				get_set_kernel_context_routine, 
				NULL, 
				NULL, 
				KernelMode, 
				NULL);
		insert_queue_apc(&ctx_context.apc, (PVOID)1, NULL, IO_NO_INCREMENT);

		resume_thread(&thread->tcb);
		status = wait_for_single_object(&ctx_context.event,
				0,
				KernelMode,
				FALSE,
				NULL);
	}
	deref_object(thread);

	if (copy_to_user(ThreadContext, &ctx_context.context, sizeof(CONTEXT)))
		status = STATUS_UNSUCCESSFUL;

	ktrace("NtGetContextThread return %x\n", status);
	return status;
}
EXPORT_SYMBOL(NtGetContextThread);

NTSTATUS SERVICECALL
NtSetContextThread(IN HANDLE   ThreadHandle,
		IN PCONTEXT ThreadContext)
{
	struct ethread * thread;
	GET_SET_CTX_CONTEXT	ctx_context;
	NTSTATUS status = STATUS_SUCCESS;

	ktrace("NtSetContextThread %p, %p\n",ThreadHandle, ThreadContext);

	if (copy_from_user(&ctx_context.context, ThreadContext, sizeof(CONTEXT)))
		return STATUS_UNSUCCESSFUL;

	status = ref_object_by_handle(ThreadHandle,
			THREAD_ALL_ACCESS,
			thread_object_type,
			KernelMode,
			(PVOID *)&thread,
			NULL);
	if (!NT_SUCCESS(status)) {
		return status;
	}


	if (thread == get_current_ethread()) {
		context_to_trapframe(&ctx_context.context, thread);
	} else {
		event_init(&ctx_context.event, notification_event, 0);

		apc_init(&ctx_context.apc, 
				&thread->tcb, 
				OriginalApcEnvironment, 
				get_set_kernel_context_routine, 
				NULL, 
				NULL, 
				KernelMode, 
				NULL);
		insert_queue_apc(&ctx_context.apc, (PVOID)0, NULL, IO_NO_INCREMENT);

		resume_thread(&thread->tcb);
		wait_for_single_object(&ctx_context.event,
				0,
				KernelMode,
				FALSE,
				NULL);
	}

	deref_object(thread);

	ktrace("NtSetContextThread return %x\n", status);
	return status;
}
EXPORT_SYMBOL(NtSetContextThread);

/* stubs from Wine server */
int thread_queue_apc( struct w32thread *thread, struct object *owner, const apc_call_t *call_data )
{
	return 0;
}

void thread_cancel_apc( struct w32thread *thread, struct object *owner, enum apc_type type )
{
}

struct w32thread *get_thread_from_id( thread_id_t id )
{
	return NULL;
}

int thread_get_inflight_fd( struct w32thread *thread, int client )
{
	return 0;
}

void send_client_fd( struct w32process* proc, unsigned int fd, void* hndl)
{
}

#endif

